
/*!
 * descent - parser
 * Copyright (c) 2011 LearnBoost <tj@learnboost.com>
 * MIT Licensed
 */

/**
 * Module dependencies.
 */

var scan = require('./lexer');

/**
 * Parse the given `str` returning a parse tree
 * represented by nested arrays.
 *
 * @param {String} str
 * @return {Array}
 * @api private
 */

module.exports = function(str) {
  var toks = scan(str)
    , tok;

  return stmts();

  /**
   * Throw `msg`.
   */

  function error(msg) {
    throw new Error(msg);
  }

  /**
   * Next token.
   */

  function next() {
    return toks.shift();
  }

  /**
   * Lookahead `n` or 1.
   */

  function peek(n){
    n = n || 1;
    return toks[--n] || ['eof'];
  }

  /**
   * Check if LA is `type`.
   */

  function is(type) {
    return type == peek()[0];
  }

  /**
   * newline*
   */

  function whitespace(){
    while (is('newline')) next();
  }

  /**
   * Expect and return `type` or throw.
   */

  function expect(type) {
    if (is(type)) return next();
    throw new Error('expected ' + type + ', got ' + peek()[0]);
  }

  /**
   * Accept and return `type`.
   */

  function accept(type) {
    if (is(type)) return next();
  }

  /**
   * stmts = ws (stmt ws)*
   */

  function stmts() {
    var stmts = [];
    whitespace();
    do {
      stmts.push(stmt() || error('invalid statement, got ' + peek()[0]));
      whitespace();
    } while (toks.length);
    return stmts;
  }

  /**
   * stmt = rule
   */

  function stmt() {
    switch (peek()[0]) {
      case 'rule': return ruleDef();
      case 'action': return prologue();
    }
  }

  /**
   * action
   */

   function prologue() {
     return ['prologue', next()[1]];
   }

  /**
   * rule-def = rule-name ws ('=' | ':') ws expr
   */

  function ruleDef() {
    tok = expect('rule');
    whitespace();
    accept(':') || expect('=');
    whitespace();
    node = ['rule-def', tok[1], expr()];
    expect(';');
    return node;
  }

  /**
   * expr = sequence ('|' sequence)?
   */

  function expr() {
    var node = ['expr', sequence()];
    if (accept('|')) return ['alt', node, expr()];
    return node;
  }

  /**
   * sequence = suffix+
   */

  function sequence() {
    var seq = []
      , node;
    while (node = suffix()) seq.push(node);
    return seq;
  }

  /**
   * suffix = primary ('*' | '+' | '?')?
   */

  function suffix() {
    var node = value();
    if (accept('*')) node = ['any', node];
    if (accept('+')) node = ['many', node];
    if (accept('?')) node = ['optional', node];
    whitespace();
    return node;
  }

  /**
   * value = primitive ws
   */

   function value() {
     var node = primitive();
     whitespace();
     return node;
   }

  /**
   * primitive = 
   *      '.'
   *    | string
   *    | regexp
   *    | rule (':' rule)?
   *    | group
   *    | action
   *    | '&' action
   *    | '!' primitive
   *    | '<' expr '>';
   */

  function primitive(){
    switch (peek()[0]) {
      case '!':
        next();
        return ['not', primitive()];
      case '.':
        next();
        return ['dot'];
      case 'string':
      case 'regexp':
      case 'action':
        return next();
      case 'rule':
        var node = next();
        if (accept(':')) {
          node = ['named-rule', node, expect('rule')];
        }
        return node;
      case '(':
        return group();
      case '<':
        return capture();
      case '&':
        next()
        return ['expr-predicate', expect('action')];
      case 'eof':
        error('unterminated rule "' + tok[1] + '", add trailing ";"');
    }
  }

  /**
   * capture = '<' ws expr '>'
   */

  function capture() {
    expect('<');
    whitespace();
    var node = expr();
    expect('>');
    return ['capture', node];
  }

  /**
   * group = '(' ws expr ')'
   */

  function group() {
    expect('(');
    whitespace();
    var node = expr();
    expect(')');
    return ['group', node];
  }
};